/*
 * ipxip.c
 */

#include <fcntl.h>
#include <errno.h>
#include "ipxip.h"

int fd_eth;
int protocol;
int frame_max = 	2000;			/* Max size for packet */
char *if_eth = 		"eth0";			/* name of ethernet interface */
int MonitorPackets = 	1;			/* monitor by default */
int timeout = 		2;
int qlength = 		256;
int sendUDP = 		1;

void SigHandlet(int);
int main(int argc, char *argv[]);
void Usage(void);
void Activate(int,int);
void IfSetFlag(int sock, char *ifname, short flag);
void IfClearFlag(int sock, char *ifname, short flag);
void get8022_addr(unsigned char *buf, int o, char *str);
unsigned short get8022_us(unsigned char *buf, int offset);
int OpenUDP_remote(char *hostname, int port);
int OpenUDP_local(int port);
int DropPacket(unsigned char *data, int len);
void ReportStats();

void
SigHandler(int sig)
	{
	IfClearFlag(fd_eth,if_eth, IFF_PROMISC);
	printf("Exit\n");
	exit(0);
	}

main(int argc, char *argv[])
	{
	char *pstr,*hostname;
	int port,udp_in,udp_out,i;

	pstr = "802.2";
	protocol = ETH_P_802_2;

	if(getuid())
	   {
	   fprintf(stderr, "must be superuser\n");
	   exit(1);
	   }

	fprintf(stderr,"IPX/IP 1.0 by Anthony Wesley\n");

	/* Process command-line args */
	for(i=1; i<argc && argv[i][0]=='-'; ++i) switch(argv[i][1])
		{
		case 'f':
		   /* Selected frame types - from linux/if_ether.h */
		   pstr = argv[++i]; protocol = -1;
		   if (!strcmp(pstr,"802.2")) protocol=ETH_P_802_2;
		   if (!strcmp(pstr,"802.3")) protocol=ETH_P_802_3;
		   if (!strcmp(pstr,"etherII")) protocol=ETH_P_IPX;
		   if (!strcmp(pstr,"echo")) protocol=ETH_P_ECHO;
		   if (!strcmp(pstr,"x25")) protocol=ETH_P_X25;
		   if (!strcmp(pstr,"arp")) protocol=ETH_P_ARP;
		   if (!strcmp(pstr,"atalk")) protocol=ETH_P_ATALK;
		   if (!strcmp(pstr,"aarp")) protocol=ETH_P_AARP;
		   if (!strcmp(pstr,"all")) protocol=ETH_P_ALL;

		   if (protocol == -1)
			{
			printf("Unknown protocol [%s]\n",pstr);
			exit(1);
			}
		   break;
		case 's':
		   frame_max = atoi(argv[++i]); break;
		case 'r': MonitorPackets = 0; break;
		case 'i': if_eth = argv[++i]; break;
		case 't': timeout = atoi(argv[++i]); break;
		case 'q': qlength = atoi(argv[++i]); break;
		case 'd': sendUDP=0; break;
		}

	if (i+2 != argc)
	   {
	   Usage();
	   exit(1);
	   }

	hostname = argv[i];
	port = atoi(argv[i+1]);

	printf("Bridging frames of type %s from local interface %s to %s:%d\n",
		pstr,if_eth,hostname,port);
	printf("Packet timeout = %d seconds, Queue length = %d frames\n",
		timeout,qlength);
	printf("  Progressive duplicate dropping: %s\n",
		MonitorPackets ? "yes" : "no");
	if (!sendUDP) printf("Test mode only - no UDP packets will be sent\n");

	/* Open output and input sockets for UDP */
	udp_out = OpenUDP_remote(hostname,port); if (udp_out<0) exit(1);
	udp_in = OpenUDP_local(port); if (udp_in<0) exit(1);

	signal(SIGINT, SigHandler);
	Activate(udp_out,udp_in);

	return 0;
	}

void
Usage()
	{
	printf("Usage: ipxip [-f fr] [-s sz] [-r] [-i if] [-t tm] [-q qs] [-u] hostname port\n\n");
	printf("fr = Ethernet frame type to bridge. Default is 802.2.\n");
	printf("     Known values are:\n");
	printf("         802.2 802.3 etherII echo x25 arp atalk aarp all\n");
	printf("sz = Maximum size frame to transmit via UDP. Default 2000 bytes\n");
	printf("-r = Reliable mode (do not monitor and drop retry-packets)\n");
	printf("     (default is to monitor and progressively drop duplicate packets)\n");
	printf("if = Network interface to read from (default is eth0)\n");
	printf("tm = Packet timeout. Do not queue packets older than tm seconds.\n");
	printf("qs = Queue size.\n");
	printf("-d = Debug mode. No UDP packets will be sent.\n");
	}

void
Activate(int sout,int sin)
	{
	fd_set	sds;
	int	max,no;
	struct sockaddr_in from;
	struct sockaddr dst;
	unsigned char *buf;
	size_t fromlen;
	int len;
	time_t now,last;

	/*
	** Allocate space for a frame
	*/
	buf = malloc(frame_max);
	if (buf==NULL)
	   {
	   fprintf(stderr,"Malloc() error.\n");
	   exit(1);
	   }

	/*
	** Open a local packet-oriented socket to interface
	** with the ethernet interface
	*/
	if ((fd_eth = socket(AF_INET, SOCK_PACKET, htons(protocol)))<0)
	   {
	   perror("Error opening local ethernet socket\n");
	   exit(1);
	   }

	IfSetFlag(fd_eth,if_eth, IFF_PROMISC);

	/* Set up a socket address to write back to ethernet */
	dst.sa_family = AF_INET;
	strcpy(dst.sa_data,if_eth);

	fromlen = sizeof(from);

	last = time(NULL);

	while(1)
	   {
	   if (time(NULL) - last > 10)
		{
		last = time(NULL);
		ReportStats();
		}

	   FD_ZERO(&sds);
	   FD_SET(fd_eth, &sds);
	   FD_SET(sin,&sds);
	   max = MAX(fd_eth,sin);

	   do no = select(max + 1, &sds, NULL, NULL, NULL);
	   while(no < 0 && errno == EINTR);

	   if(FD_ISSET(fd_eth,&sds))
		{
		/*
		** Read a packet from the ethernet fd and write it
		** out to the UDP fd.
		*/
		len = recvfrom(fd_eth, buf, frame_max, 0,
			(struct sockaddr *) &from, &fromlen);

		if (MonitorPackets==0 || DropPacket(buf,len) == 0)
		   if (sendUDP) write(sout,buf,len);
		}

	   if(FD_ISSET(sin,&sds))
		{
		/*
		** Read a packet from the UDP fd and write it back
		** out onto the local ethernet
		*/
		len = recvfrom(sin, buf, frame_max, 0, 
			(struct sockaddr *)&from, &fromlen);
		sendto(fd_eth,buf,len,0,&dst,sizeof(dst));
		}
	   }
	}

void
IfSetFlag(int sock, char *ifname, short flag)
	{
	struct ifreq ifr;

	strcpy(ifr.ifr_name, ifname);
	if(ioctl(sock, SIOCGIFFLAGS, &ifr) < 0) {
		perror("SIOCGIFFLAGS");
		return;
	}
	ifr.ifr_flags |= flag;
	if(ioctl(sock, SIOCSIFFLAGS, &ifr) < 0)
		perror("SIOCSIFFLAGS");
	}

void
IfClearFlag(int sock, char *ifname, short flag)
{
	struct ifreq ifr;

	strcpy(ifr.ifr_name, ifname);
	if(ioctl(sock, SIOCGIFFLAGS, &ifr) < 0) {
		perror("SIOCGIFFLAGS");
		return;
	}
	ifr.ifr_flags &= ~flag;
	if(ioctl(sock, SIOCSIFFLAGS, &ifr) < 0)
		perror("SIOCSIFFLAGS");
}

/************************************************************************
************************************************************************/

int
OpenUDP_local(int port)
	{
	int s,ns,alen;
	struct sockaddr_in my;

	s = socket(AF_INET, SOCK_DGRAM,0);
	if (s<0)
	   {
	   fprintf(stderr,"Could not create UDP socket\n");
	   exit(1);
	   }

	/* Accept any connections on the nominated port */
	my.sin_family = AF_INET;
	my.sin_addr.s_addr = INADDR_ANY;
	my.sin_port = htons(port);

	if (bind(s,(struct sockaddr *)&my,sizeof(my)) != 0)
	   {
	   fprintf(stderr,"Could not bind socket to address\n");
	   exit(1);
	   }
	fprintf(stderr,"Socket bound ok\n"); fflush(stdout);

	return s;
	}

/*
** This function opens a UDP connection-based socket to the named
** remote host/port and returns the file descriptor for use with
** read/write/select
*/

int
OpenUDP_remote(char *hostname, int port)
	{
	int s,n;
	struct sockaddr_in remote;
	struct in_addr in;
	struct hostent *he;
	int addrlen;

	/* Create a datagram socket */
	s = socket(AF_INET,SOCK_DGRAM,0);
	if (s<0)
	   {
	   fprintf(stderr,"Could not create UDP socket\n");
	   exit(1);
	   }

	/* convert hostname from ascii -> binary */
	if (inet_aton(hostname,&in) == 0)
	   {
	   /* Might be a hostname - do lookup */
	   he = gethostbyname(hostname);
	   if (he==NULL)
		{
		fprintf(stderr,"Unknown host %s\n",hostname);
	   	exit(1);
		}
	   memcpy((char *)&in,(char *)he->h_addr_list[0],sizeof(in));
	   }

	remote.sin_family = AF_INET;
	remote.sin_addr = in;
	remote.sin_port = htons(port);

	/* Connect the socket */
	if (connect(s, (struct sockaddr *) &remote, sizeof(remote)) != 0)
	   {
	   fprintf(stderr,"Could not connect to %s\n",hostname);
	   exit(1);
	   }

	return s;
	}

/**************************************************************************
**************************************************************************/


void get8022_addr(unsigned char *buf, int o, char *str)
	{
	sprintf(str,"%02x:%02x:%02x:%02x:%02x:%02x",
		buf[o],buf[o+1],buf[o+2],buf[o+3],buf[o+4],buf[o+5]);
	}

unsigned short
get8022_us(unsigned char *buf, int o)
	{
	unsigned short l;

	l = buf[o]; l<<=8; l+=buf[o+1];
	return l;
	}

/****************************************************************************
****************************************************************************/

#define INCREMENT(x)  (++x == qlength ? x=0: x)
struct lentry
	{
	time_t tm;
	unsigned int chk;
	};

static int TotalPackets = 	0;
static int PacketsDropped = 	0;
static int Duplicates = 	0;
static int MaxQlength = 	0;
static int Qcount = 		0;
static int Qentries = 		0;

int
DropPacket(unsigned char *data, int len)
	{
	static struct lentry *buf;
	static int head=0,tail=0,lookup[256],linit=0;
	int i,count;
	unsigned int r,chk,n;
	time_t now;
	
	if (linit==0)
	   {
	   linit=1;
	   for(i=0; i<256; ++i) lookup[i] = rand();

	   buf = malloc(qlength * sizeof(struct lentry));
	   if (buf==NULL)
		{
		printf("Malloc() error\n");
		exit(1);
		}
	   }

	++TotalPackets;

	/* calculate checksum */
	for(i=0,chk=len; i<len; ++i)
	   chk += lookup[(chk&255) ^ data[i]];

	/* Remove Expired packets at head of list */
	now = time(NULL);
	while(head != tail && now - buf[head].tm >= timeout)
	   INCREMENT(head);
	   
	/* If the list is full, lose the first entry to make room */
	i=tail; INCREMENT(i);
	if (i==head) INCREMENT(head);

	/* Count duplicate packets in list */
	for(i=head,count=0; i != tail; INCREMENT(i))
	   if (buf[i].chk == chk) ++count;

	/* Append our entry to the end */
	buf[tail].tm = now;
	buf[tail].chk = chk;
	INCREMENT(tail);

	i = tail - head; if (i<0) i+= qlength;
	if (i>MaxQlength) MaxQlength = i;

	++Qcount; Qentries += i;

	if (count == 0) return(0);
	++Duplicates;

	/*
	** This algorithm provides a rapidly decreasing probability for
	** sending packets which are already in the Queue, down to a
	** minimum probability if 1/qlength.
	*/

	r = rand() % qlength;    /* random number 0 <= r <= qlength */
	n = qlength >> count;	 /* more duplicates -> smaller window n */
	if (r <= n ) return 0;   /* if r is in our window, then ok */

	++PacketsDropped;
	return(1);
	}

void
ReportStats()
	{
	printf("Total packets	 :%d (%d duplicates, %d dropped)\n",
		TotalPackets, Duplicates, PacketsDropped);
	printf("Maximum Queue length: %d (%d %%), Average Queue length = %d (%d %%)\n",
	   MaxQlength, MaxQlength * 100 / qlength, Qentries / (Qcount+1),
	   (Qentries/(Qcount+1)) * 100 / qlength);
	putchar('\n'); fflush(stdout);

	TotalPackets = Duplicates = PacketsDropped = 0;
	MaxQlength = Qentries = Qcount = 0;
	}
