/*
 * ipxip.c
 */

#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <unistd.h>
#include <limits.h>
#include <signal.h>
#include <string.h>
#include <netdb.h>
#include <errno.h>
#include <endian.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <arpa/nameser.h>
#include <arpa/inet.h>
#include <linux/if_ether.h>

#define MAX_PACKET_LENGTH 1600
#define MAX_CLIENTS 16
#define QLENGTH 512
#define STATS_TIME 2
#define HASH_TABLE_SIZE (0x10000)

enum {False=0, True=1};
#define BOOL int

static int eth_fd = -1;
static char *if_eth = "eth0";	/* name of ethernet interface */
static int tcp_fd = -1; /* fd of peer */

struct history {
	int stamp;
	short length;
	char data[MAX_PACKET_LENGTH];
};

/* these arrays hold the packet history used to detect/handle
   duplicates. The length is in bytes and is sent in 
   network byte order */
static struct history local_history[QLENGTH];
static struct history remote_history[QLENGTH];

struct stats {
	int remote_pcnt; /* packet count from peer */
	int local_pcnt; /* local packet count */
	int local_wptr; /* write pointer - the number of packets
			  actually sent down the link */
	int local_dup; /* number of local duplicates */
	int remote_dup; /* number of local duplicates */
	int max_tdiff; /* max time difference */
	int tot_tdiff; /* total time diff over the packets */
	int byte_count; /* byte transferred */
	int max_queue; /* maximum backlog */
	int dropped;   /* num dropped */
};


struct stats stats;
struct stats last_stats;

/* the checksums of the local packets */
static unsigned checksums[QLENGTH];
static unsigned hash_table[HASH_TABLE_SIZE];

#define CUTOFF_START 2
#define CUTOFF_END 20

#ifndef MAX
#define MAX(a,b) ((a)<(b)?(b):(a))
#endif

/* a simple xor based checksum routine - note that it doesn't need to be 
   good */
static unsigned csum(char *data, int len)
{
	unsigned *idata = (unsigned *)data;
	unsigned sum = 0;
	int rot=0;

	len /= sizeof(sum);

	while (len--) {
		sum ^= (idata[0] << rot) | (idata[0] >> (32-rot));
		rot++;
		idata++;
	}

	return sum;
}


/* a relative time in hundreds of microseconds */
static int timestamp(void)
{
	static struct timeval first;
	static int initialised;
	struct timeval tv;

	if (!initialised) {
		initialised = 1;
		gettimeofday(&first, NULL);		
	}
	gettimeofday(&tv, NULL);

	return (tv.tv_sec - first.tv_sec)*10000  + 
		((tv.tv_usec - first.tv_usec)/100);
}


/* find a matching packet in the history */
static void insert_csum(int pnum, unsigned sum)
{
	unsigned hash = sum % (HASH_TABLE_SIZE);

	hash_table[hash] = pnum;
}

/* find a matching packet in the history */
static int find_csum(unsigned sum, char *data, int len)
{
	unsigned hash = sum % (HASH_TABLE_SIZE);
	int pnum;

	pnum = hash_table[hash];

	if (local_history[pnum].length == len &&
	    memcmp(local_history[pnum].data, data, len) == 0) {
		return pnum;
	}

#if DEBUG
	{
		int i;
		for (i=0;i<QLENGTH;i++)
			if (local_history[i].length == len &&
			    memcmp(local_history[i].data, data, len) == 0) {
				printf("match at %d\n", i);
				break;
			}
	}
#endif

	return -1;
}


/* write data to a socket - taking account of the fact that it may
   occur in pieces */
static int sock_write(int fd, char *data, int len)
{
	int total = 0;

	while (len) {
		int s = write(fd, data, len);
		if (s <= 0) {
			return -1;
		}
		data += s;
		len -= s;
		total += s;
	}
	return total;
}

/* read data from a socket - taking account of the fact that it may
   occur in pieces */
static int sock_read(int fd, char *data, int len)
{
	int total = 0;

	while (len) {
		int s = read(fd, data, len);
		if (s <= 0) {
			return -1;
		}
		data += s;
		len -= s;
		total += s;
	}
	return total;
}


/* open a tcp connection to a particular host and port and return a file 
   descriptor for it */
static int open_tcp_out(char *hostname, int port)
{
	struct sockaddr_in remote;
	int fd;
	struct in_addr in;
	struct hostent *he;

	/* convert hostname from ascii -> binary */
	if (inet_aton(hostname,&in) == 0) {
		/* Might be a hostname - do lookup */
		he = gethostbyname(hostname);
		if (he==NULL) {
			fprintf(stderr,"Unknown host %s\n",hostname);
			return -1;
		}
		memcpy((char *)&in,(char *)he->h_addr_list[0],sizeof(in));
	}

	/* create a socket to write to */
	fd = socket(PF_INET, SOCK_STREAM, 0);
	if (fd == -1) {
		perror("socket");
		return -1;
	}
  
	remote.sin_family = AF_INET;
	remote.sin_addr = in;
	remote.sin_port = htons(port);

	if (connect(fd,(struct sockaddr *)&remote,sizeof(remote))) {
		perror("connect");
		return -1;
	}

	return fd;
}

int open_tcp_in(int port)
{
	struct hostent *hp;
	struct sockaddr_in sock;
	char host_name[1024];
	int fd, fd2;
	struct sockaddr addr;
	size_t in_addrlen = sizeof(addr);

	/* get my host name */
	if (gethostname(host_name, MAXHOSTNAMELEN) == -1) {
		perror("gethostname");
		return -1;
	}

	/* get host info */
	if ((hp = gethostbyname(host_name)) == 0) {
		perror("gethostbyname");
		return -1;
	}
  
	bzero((char *)&sock,sizeof(sock));
	memcpy((char *)&sock.sin_addr,(char *)hp->h_addr, hp->h_length);
	sock.sin_port = htons(port);
	sock.sin_family = hp->h_addrtype;
	sock.sin_addr.s_addr = INADDR_ANY;

	fd = socket(hp->h_addrtype, SOCK_STREAM, 0);
	if (fd == -1) {
		perror("socket");
		return -1;
	}

	{
		int one=1;
		setsockopt(fd,SOL_SOCKET,SO_REUSEADDR,(char *)&one,
			   sizeof(one));
	}

	/* now we've got a socket - we need to bind it */
	if (bind(fd, (struct sockaddr * ) &sock,sizeof(sock)) < 0) {
		perror("bind");
		return -1;
	}

	if (listen(fd, 2) == -1) {
		perror("listen");
		return -1;
	}

	printf("waiting for connection\n");
	fd2 = accept(fd,&addr,&in_addrlen);
	close(fd);
	if (fd2 == -1) {
		perror("accept");
		return -1;
	}

	return fd2;
}



enum SOCK_OPT_TYPES {OPT_BOOL,OPT_INT,OPT_ON};

struct
{
  char *name;
  int level;
  int option;
  int value;
  int opttype;
} socket_options[] = {
  {"SO_KEEPALIVE",      SOL_SOCKET,    SO_KEEPALIVE,    0,                 OPT_BOOL},
  {"SO_REUSEADDR",      SOL_SOCKET,    SO_REUSEADDR,    0,                 OPT_BOOL},
  {"SO_BROADCAST",      SOL_SOCKET,    SO_BROADCAST,    0,                 OPT_BOOL},
#ifdef TCP_NODELAY
  {"TCP_NODELAY",       IPPROTO_TCP,   TCP_NODELAY,     0,                 OPT_BOOL},
#endif
#ifdef IPTOS_LOWDELAY
  {"IPTOS_LOWDELAY",    IPPROTO_IP,    IP_TOS,          IPTOS_LOWDELAY,    OPT_ON},
#endif
#ifdef IPTOS_THROUGHPUT
  {"IPTOS_THROUGHPUT",  IPPROTO_IP,    IP_TOS,          IPTOS_THROUGHPUT,  OPT_ON},
#endif
#ifdef SO_SNDBUF
  {"SO_SNDBUF",         SOL_SOCKET,    SO_SNDBUF,       0,                 OPT_INT},
#endif
#ifdef SO_RCVBUF
  {"SO_RCVBUF",         SOL_SOCKET,    SO_RCVBUF,       0,                 OPT_INT},
#endif
#ifdef SO_SNDLOWAT
  {"SO_SNDLOWAT",       SOL_SOCKET,    SO_SNDLOWAT,     0,                 OPT_INT},
#endif
#ifdef SO_RCVLOWAT
  {"SO_RCVLOWAT",       SOL_SOCKET,    SO_RCVLOWAT,     0,                 OPT_INT},
#endif
#ifdef SO_SNDTIMEO
  {"SO_SNDTIMEO",       SOL_SOCKET,    SO_SNDTIMEO,     0,                 OPT_INT},
#endif
#ifdef SO_RCVTIMEO
  {"SO_RCVTIMEO",       SOL_SOCKET,    SO_RCVTIMEO,     0,                 OPT_INT},
#endif
  {NULL,0,0,0,0}};


BOOL next_token(char **ptr,char *buff,char *sep)
{
  char *s;
  BOOL quoted;
  static char *last_ptr=NULL;

  if (!ptr) ptr = &last_ptr;
  if (!ptr) return(False);

  s = *ptr;

  /* default to simple separators */
  if (!sep) sep = " \t\n\r";

  /* find the first non sep char */
  while(*s && strchr(sep,*s)) s++;

  /* nothing left? */
  if (! *s) return(False);

  /* copy over the token */
  for (quoted = False; *s && (quoted || !strchr(sep,*s)); s++)
    {
      if (*s == '\"') 
	quoted = !quoted;
      else
	*buff++ = *s;
    }

  *ptr = (*s) ? s+1 : s;  
  *buff = 0;
  last_ptr = *ptr;

  return(True);
}
	

/****************************************************************************
set user socket options
****************************************************************************/
static void set_socket_options(int fd, char *options)
{
  char tok[100];

  while (next_token(&options,tok," \t,"))
    {
      int ret=0,i;
      int value = 1;
      char *p;
      BOOL got_value = False;

      if ((p = strchr(tok,'=')))
	{
	  *p = 0;
	  value = atoi(p+1);
	  got_value = True;
	}

      for (i=0;socket_options[i].name;i++)
	if (strcmp(socket_options[i].name,tok) == 0)
	  break;

      if (!socket_options[i].name)
	{
	  printf("Unknown socket option %s\n",tok);
	  continue;
	}

      switch (socket_options[i].opttype)
	{
	case OPT_BOOL:
	case OPT_INT:
	  ret = setsockopt(fd,socket_options[i].level,
			   socket_options[i].option,(char *)&value,sizeof(int));
	  break;

	case OPT_ON:
	  if (got_value)
	    printf("syntax error - %s does not take a value\n",tok);

	  {
	    int on = socket_options[i].value;
	    ret = setsockopt(fd,socket_options[i].level,
			     socket_options[i].option,(char *)&on,sizeof(int));
	  }
	  break;	  
	}
      
      if (ret != 0)
	printf("Failed to set socket option %s\n",tok);
    }
}



/* close the ethernet fd and set it back to non-promiscuous */
static void eth_close(int fd, char *interface)
{
	struct ifreq ifr;

	strcpy(ifr.ifr_name, interface);
	if(ioctl(fd, SIOCGIFFLAGS, &ifr) < 0) {
		perror("SIOCGIFFLAGS");
		return;
	}

	ifr.ifr_flags &= ~IFF_PROMISC;
	if(ioctl(fd, SIOCSIFFLAGS, &ifr) < 0) {
		perror("SIOCSIFFLAGS");
		return;
	}

	close(fd);
}


/* open an ethernet socket for the requested protocol and
 return a file descriptor that can be used for reading/writing packets
 to that interface
 
 pstr specifies what packet types to accept
 */
static int eth_open(char *interface, char *pstr)
{
	int fd;
	int protocol = -1;
	struct ifreq ifr;

	if (!strcmp(pstr,"802.2")) protocol=ETH_P_802_2;
	if (!strcmp(pstr,"802.3")) protocol=ETH_P_802_3;
	if (!strcmp(pstr,"etherII")) protocol=ETH_P_IPX;
	if (!strcmp(pstr,"echo")) protocol=ETH_P_ECHO;
	if (!strcmp(pstr,"x25")) protocol=ETH_P_X25;
	if (!strcmp(pstr,"arp")) protocol=ETH_P_ARP;
	if (!strcmp(pstr,"atalk")) protocol=ETH_P_ATALK;
	if (!strcmp(pstr,"aarp")) protocol=ETH_P_AARP;
	if (!strcmp(pstr,"all")) protocol=ETH_P_ALL;

	if (protocol == -1) {
		printf("Unknown protocol [%s]\n",pstr);
		return -1;
	}

	if ((fd = socket(AF_INET, SOCK_PACKET, htons(protocol)))<0) {
		perror("Error opening local ethernet socket\n");
		return -1;
	}

	strcpy(ifr.ifr_name, interface);
	if(ioctl(fd, SIOCGIFFLAGS, &ifr) < 0) {
		perror("SIOCGIFFLAGS");
		return -1;
	}

	ifr.ifr_flags |= IFF_PROMISC;
	if(ioctl(fd, SIOCSIFFLAGS, &ifr) < 0) {
		perror("SIOCSIFFLAGS");
		return -1;
	}

	return fd;
}


static int eth_write(int fd, char *data, int len)
{
	struct sockaddr dst;

	dst.sa_family = AF_INET;
	strcpy(dst.sa_data,if_eth);

	return sendto(fd,data,len,0,&dst,sizeof(dst));
}

static int eth_read(int fd, char *buf)
{
	struct sockaddr_in from;
	size_t fromlen = sizeof(from);

	return recvfrom(fd, buf, MAX_PACKET_LENGTH, 0,
			(struct sockaddr *) &from, &fromlen);
}

static void show_stats(void)
{
	printf("local=%d/%d remote=%d ldup=%d rdup=%d tmax=%d tavg=%d bytes=%d qmax=%d drop=%d\n",
	       stats.local_wptr - last_stats.local_wptr, 
	       stats.local_pcnt - last_stats.local_pcnt, 
	       stats.remote_pcnt - last_stats.remote_pcnt,
	       stats.local_dup - last_stats.local_dup, 
	       stats.remote_dup - last_stats.remote_dup,
	       stats.max_tdiff,
	       (stats.tot_tdiff - last_stats.tot_tdiff) / 
	       (1 + stats.local_wptr - last_stats.local_wptr),
	       stats.byte_count - last_stats.byte_count,
	       stats.max_queue,
	       stats.dropped - last_stats.dropped);

	stats.max_tdiff = 0;
	stats.max_queue = 0;
}

static void Usage(void)
{
	printf("Usage: \nipxip [-f fr] [-i if] hostname port\n\n");
	printf("ipxip [-f fr] [-i if] -S port\n\n");
	printf("fr = Ethernet frame type to bridge. Default is 802.2.\n");
	printf("     Known values are:\n");
	printf("         802.2 802.3 etherII echo x25 arp atalk aarp all\n");
	printf("if = Network interface to read from (default is eth0)\n");
}


void sig_int(int sig)
{
	eth_close(eth_fd, if_eth);
	printf("Exit\n");
	exit(0);
}


static int drop_packet(int len)
{
	int backlog = stats.local_pcnt - stats.local_wptr;
	int r;

	if (backlog < CUTOFF_START) return 0;

	r = random() % (CUTOFF_END - CUTOFF_START);

	if (backlog > CUTOFF_END) {
		return r != 0;
	}

	if (r < (backlog - CUTOFF_START)) {
		stats.dropped++;
		return 1;
	}

	return 0;
}

int main(int argc, char *argv[])
{
	char *hostname;
	int i;
	int port;
	char *pstr = "802.2";
	int server=0;
	time_t lastt=0, t;
 	char *socket_ops = "SO_SNDBUF=4096";

	if(getuid()) {
		fprintf(stderr, "must be superuser\n");
		exit(1);
	}

	fprintf(stderr,"IPX/IP 1.0 by Anthony Wesley and Andrew Tridgell\n");

	/* Process command-line args */
	for(i=1; i<argc && argv[i][0]=='-'; ++i) switch(argv[i][1])
		{
		case 'f':
		   /* Selected frame types - from linux/if_ether.h */
		   pstr = argv[++i];
		   break;
		case 'S': server = 1; break;
		case 'i': if_eth = argv[++i]; break;
		case 'O': socket_ops = argv[++i]; break;
		}

	eth_fd = eth_open(if_eth, pstr);
	if (eth_fd == -1) {
		exit(1);
	}

	signal(SIGINT, sig_int);

	if (server) {
		if (i+1 != argc) {
			Usage();
			exit(1);
		}

		port = atoi(argv[i]);

		tcp_fd = open_tcp_in(port);
		if (tcp_fd == -1) {
			exit(1);
		}
	} else {
		if (i+2 != argc) {
			Usage();
			exit(1);
		}

		hostname = argv[i];
		port = atoi(argv[i+1]);

		tcp_fd = open_tcp_out(hostname, port);
		if (tcp_fd == -1) {
			exit(1);
		}
	}

	set_socket_options(tcp_fd, socket_ops);

	printf("connected\n");

	while (1) {
		fd_set	rfd, wfd;
		int no, pnum, pnum2;
		struct timeval tval;

		FD_ZERO(&rfd);
		FD_ZERO(&wfd);
		FD_SET(eth_fd, &rfd);
		FD_SET(tcp_fd, &rfd);
		if (stats.local_pcnt != stats.local_wptr)
			FD_SET(tcp_fd, &wfd);
		
		tval.tv_sec = 1;
		tval.tv_usec = 0;

		t = time(NULL);
		if (t > lastt + STATS_TIME) {
			show_stats();
			lastt = t;
			last_stats = stats;
		}

		no = select(32, &rfd, &wfd, NULL, &tval);

		if (no == -1 && errno != EAGAIN) {
			printf("select error\n");
			break;
		}

		if (no == 0) continue;

		/* is anything happening on the ethernet? */
		if (FD_ISSET(eth_fd, &rfd)) {
			unsigned sum;
			int len;

			pnum = stats.local_pcnt % QLENGTH;
			local_history[pnum].length = 0;

			len = eth_read(eth_fd, local_history[pnum].data);

			if (len > 0 && !drop_packet(len)) {
				local_history[pnum].stamp = timestamp();
				sum = csum(local_history[pnum].data, len); 

				pnum2 = find_csum(sum, 
						  local_history[pnum].data,
						  len);

				if (pnum2 != -1) {
#if DEBUG
					printf("dup at %d (was %d) of len %d\n",
					       pnum, pnum2, 
					       len);
#endif
					/* found a duplicate */
					stats.local_dup++;
					local_history[pnum].length = -pnum2;
					checksums[pnum] = 0;
				} else {
					checksums[pnum] = sum;
					local_history[pnum].length = len;
					insert_csum(pnum, sum);
				}

				stats.local_pcnt++;
				stats.byte_count += len;
				if (stats.local_pcnt % QLENGTH ==
				    stats.local_wptr % QLENGTH) {
					printf("ring buffer overflow!\n");
					stats.local_wptr = stats.local_pcnt;
				}
			}
		}

		/* has something arrived via tcp? */
		if (FD_ISSET(tcp_fd, &rfd)) {
			short len;

			pnum = stats.remote_pcnt % QLENGTH;

			if (sock_read(tcp_fd, (char *)&len, sizeof(len)) != 
			    sizeof(len)) {
				printf("read error - exiting\n");
				exit(1);
			}
			
			/* convert to local byte order */
			len = ntohs(len);

			if (len < -QLENGTH ||
			    len > MAX_PACKET_LENGTH) {
				printf("invalid packet length %d\n",
				       (int)len);
				break;
			}
			
			if (len <= 0) {
				/* its a duplicate marker */
				stats.remote_dup++;
				pnum = -len;
				eth_write(eth_fd, 
					  remote_history[pnum].data,
					  remote_history[pnum].length);
			} else {
				/* its real data */
				if (sock_read(tcp_fd, 
					      remote_history[pnum].data, 
					      len) != len) {
					printf("read error\n");
					break;
				}
				remote_history[pnum].length = len;
				eth_write(eth_fd, 
					  remote_history[pnum].data,
					  remote_history[pnum].length);
			}
			stats.byte_count += remote_history[pnum].length;
			stats.remote_pcnt++;
		}

		/* can we write to the socket? */
		if (stats.local_wptr != stats.local_pcnt &&
		    FD_ISSET(tcp_fd, &wfd)) {
			char buf[MAX_PACKET_LENGTH + 32];
			int buflen;
			short len;
			int tdiff;

			pnum = stats.local_wptr % QLENGTH;

			len = htons(local_history[pnum].length);
			
			memcpy(buf, (char *)&len, sizeof(len));
			buflen = sizeof(len);

			if (local_history[pnum].length > 0) {
				memcpy(buf+buflen,
				       local_history[pnum].data,
				       local_history[pnum].length);
				buflen += local_history[pnum].length;
			}

			tdiff = timestamp() - local_history[pnum].stamp;
			stats.max_tdiff = MAX(tdiff, stats.max_tdiff);
			stats.tot_tdiff += tdiff;

#if DEBUG
			if (tdiff > 1000) {
				printf("tdiff=%d t1=%d t2=%d len=%d\n",
				       tdiff, local_history[pnum].stamp,
				       timestamp(), 
				       local_history[pnum].length);
			}
#endif

			if (sock_write(tcp_fd, buf, buflen) != buflen) {
				printf("tcp write error\n");
				break;
			}

			stats.max_queue = MAX(stats.max_queue, 
					      stats.local_pcnt - 
					      stats.local_wptr);
			
			stats.local_wptr++;
		}
	}

	printf("finished\n");
	eth_close(eth_fd, if_eth);

	return 0;
}

